@@ -562,7 +562,17 @@ globus_i_gfs_log_close(void)
     if(globus_l_gfs_log_handle != NULL)
     {
         globus_logging_flush(globus_l_gfs_log_handle);
+        /*
+            NOTE: We do not destroy this handle.  At log-close time,
+            there may be several other threads that try to subsequently
+            log:
+            - Watchdog callback for data / control channels (race condition)
+            - DSI code during shutdown or threads.
+            If they try to grab the destroyed mutex, they may deadlock.
+            Since access to the pointer is not threadsafe, we cannot simply
+            set it to NULL.
         globus_logging_destroy(globus_l_gfs_log_handle);
+        */
     }
     if(globus_l_gfs_log_file != stderr && globus_l_gfs_log_file != NULL)
     {
@@ -849,7 +859,8 @@ globus_i_gfs_log_create_transfer_event_msg(
     globus_off_t                        nbytes,
     char *                              type,
     char *                              username,
-    char *                              retransmit_str)
+    char *                              retransmit_str,
+    char *                              taskid)
 {
     char *                              transfermsg;
     GlobusGFSName(globus_i_gfs_log_create_transfer_event_msg);
@@ -864,7 +875,8 @@ globus_i_gfs_log_create_transfer_event_msg(
         "streams=%d "
         "stripes=%d "
         "remoteIP=%s "
-        "type=%s"
+        "type=%s "
+        "taskid=%s"
         "%s%s",
         username,
         fname,
@@ -875,6 +887,7 @@ globus_i_gfs_log_create_transfer_event_msg(
         stripe_count,
         dest_ip,
         type,
+        taskid ? taskid : "none",
         retransmit_str ? " retrans=" : "",
         retransmit_str ? retransmit_str : "");
 
@@ -897,7 +910,8 @@ globus_i_gfs_log_transfer(
     char *                              volume,
     char *                              type,
     char *                              username,
-    char *                              retransmit_str)
+    char *                              retransmit_str,
+    char *                              taskid)
 {
     time_t                              start_time_time;
     time_t                              end_time_time;
@@ -940,7 +954,7 @@ globus_i_gfs_log_transfer(
         win_size = tcp_bs;
     }
 
-    sprintf(out_buf,
+    snprintf(out_buf, sizeof(out_buf),
         "DATE=%04d%02d%02d%02d%02d%02d.%06d "
         "HOST=%s "
         "PROG=%s "
@@ -956,7 +970,8 @@ globus_i_gfs_log_transfer(
         "STRIPES=%d "
         "DEST=[%s] "
         "TYPE=%s "
-        "CODE=%d"
+        "CODE=%d "
+        "TASKID=%s"
         "%s%s\n",
         /* end time */
         end_tm_time.tm_year + 1900,
@@ -988,9 +1003,12 @@ globus_i_gfs_log_transfer(
         dest_ip,
         type,
         code,
+        taskid ? taskid : "none",
         retransmit_str ? " retrans=" : "",
         retransmit_str ? retransmit_str : "");
 
+    out_buf[sizeof(out_buf)-1] = '\0';
+
     if(globus_l_gfs_transfer_log_file != NULL)
     {
         fwrite(out_buf, 1, strlen(out_buf), globus_l_gfs_transfer_log_file);
